using System;
using System.Collections.Generic;
using System.Text;
using System.Data;
using System.ComponentModel;
using MySql.Data.VisualStudio.Properties;
using System.Runtime.InteropServices;
using Microsoft.VisualStudio.OLE.Interop;
using MySql.Data.VisualStudio.Editors;
using System.Windows.Forms;

namespace MySql.Data.VisualStudio.DbObjects
{
    public enum InsertMethod
    {
        No, First, Last
    }

    public enum PackKeysMethod
    {
        Default, None, Full
    }

	internal class Table
	{
        internal struct TableProperties: ICustomTypeDescriptor 
        {
            private string characterSet;

            public TableProperties(DataRow row) : this()
            {
                Schema = row["TABLE_SCHEMA"].ToString();
                Name = row["TABLE_NAME"].ToString();
                Engine = row["ENGINE"].ToString();
                RowFormat = (RowFormat)Enum.Parse(typeof(RowFormat), row["ROW_FORMAT"].ToString());
                AvgRowLength = DataRowHelpers.GetValueAsUInt64(row, "AVG_ROW_LENGTH");
                AutoInc = DataRowHelpers.GetValueAsUInt64(row, "AUTO_INCREMENT");
                Comment = row["TABLE_COMMENT"].ToString();
                Collation = row["TABLE_COLLATION"].ToString();
                if (Collation != null)
                {
                    int index = Collation.IndexOf("_");
                    if (index != -1)
                        CharacterSet = Collation.Substring(0, index);
                }

                string createOpt = (string)row["CREATE_OPTIONS"];
                if (!String.IsNullOrEmpty(createOpt))
                    ParseCreateOptions(createOpt.ToLowerInvariant());
            }

            private void ParseCreateOptions(string createOptions)
            {
                string[] options = createOptions.Split(' ');
                foreach (string option in options)
                {
                    string[] parts = option.Split('=');
                    if (parts.Length != 2) continue;
                    switch (parts[0])
                    {
                        case "min_rows":
                            MinRows = UInt64.Parse(parts[1]);
                            break;
                        case "max_rows":
                            MaxRows = UInt64.Parse(parts[1]);
                            break;
                        case "checksum":
                            CheckSum = Boolean.Parse(parts[1]);
                            break;
                        case "pack_keys":
                            PackKeys = parts[1] == "1" ? PackKeysMethod.Full : PackKeysMethod.None;
                            break;
                        case "delay_key_write":
                            DelayKeyWrite = parts[1] == "1";
                            break;
                        // data directory
                        // index directory
                        // insert method
                    }
                }
            }

            #region Table options

            [Category("(Identity)")]
            [MyDescription("TableNameDesc")]
            public string Name { get; set; }

            [Category("(Identity)")]
            [MyDescription("TableSchemaDesc")]
            public string Schema { get; set; }

            [MyDescription("TableCommentDesc")]
            public string Comment { get; set; }

            [Category("Table Options")]
            [DisplayName("Character Set")]
            [TypeConverter(typeof(CharacterSetTypeConverter))]
            [RefreshProperties(RefreshProperties.All)]
            [MyDescription("TableCharSetDesc")]
            public string CharacterSet
            {
                get { return characterSet; }
                set
                {
                    if (value != characterSet)
                        Collation = String.Empty;
                    characterSet = value;
                }
            }

            [Category("Table Options")]
            [DisplayName("Collation")]
            [TypeConverter(typeof(CollationTypeConverter))]
            [MyDescription("TableCollationDesc")]
            public string Collation { get; set; }

            [Category("Table")]
            [DisplayName("Auto Increment")]
            [MyDescription("TableAutoIncStartDesc")]
            public ulong AutoInc { get; set; }

            #endregion

            #region Storage options

            [Category("Storage")]
            [DisplayName("Storage Engine")]
            [MyDescription("TableEngineDescription")]
            [TypeConverter(typeof(TableEngineTypeConverter))]
            [RefreshProperties(RefreshProperties.All)]
            public string Engine { get; set; }

            [Category("Storage")]
            [DisplayName("Data Directory")]
            [MyDescription("TableDataDirDesc")]
            public string DataDirectory { get; set; }

            [Category("Storage")]
            [DisplayName("Index Directory")]
            [MyDescription("TableIndexDirDesc")]
            public string IndexDirectory { get; set; }

            #endregion

            #region Row options

            [Category("Row")]
            [DisplayName("Row Format")]
            [DefaultValue(RowFormat.Default)]
            [MyDescription("TableRowFormatDesc")]
            public RowFormat RowFormat { get; set; }

            [Category("Row")]
            [DisplayName("Compute Checksum")]
            [MyDescription("TableCheckSumDesc")]
            [DefaultValue(false)]
            [TypeConverter(typeof(YesNoTypeConverter))]
            public bool CheckSum { get; set; }

            [Category("Row")]
            [DisplayName("Average Row Length")]
            [MyDescription("TableAvgRowLengthDesc")]
            [TypeConverter(typeof(NumericTypeConverter))]
            public ulong AvgRowLength { get; set; }

            [Category("Row")]
            [DisplayName("Minimum Rows")]
            [MyDescription("TableMinRowsDesc")]
            [TypeConverter(typeof(NumericTypeConverter))]
            public ulong MinRows { get; set; }

            [Category("Row")]
            [DisplayName("Maximum Rows")]
            [MyDescription("TableMaxRowsDesc")]
            [TypeConverter(typeof(NumericTypeConverter))]
            public UInt64 MaxRows { get; set; }

            [Category("Row")]
            [DisplayName("Pack Keys")]
            [MyDescription("TablePackKeysDesc")]
            [DefaultValue(PackKeysMethod.Default)]
            public PackKeysMethod PackKeys { get; set; }

            [Category("Row")]
            [DisplayName("Insert method")]
            [MyDescription("TableInsertMethodDesc")]
            [DefaultValue(InsertMethod.First)]
            public InsertMethod InsertMethod { get; set; }

            [Category("Row")]
            [DisplayName("Delay Key Write")]
            [MyDescription("DelayKeyWriteDesc")]
            public bool DelayKeyWrite { get; set; }

            #endregion

            #region ICustomTypeDescriptor Members

            public TypeConverter GetConverter()
            {
                return TypeDescriptor.GetConverter(this, true);
            }

            public EventDescriptorCollection GetEvents(Attribute[] attributes)
            {
                return TypeDescriptor.GetEvents(this, attributes, true);
            }

            EventDescriptorCollection System.ComponentModel.ICustomTypeDescriptor.GetEvents()
            {
                return TypeDescriptor.GetEvents(this, true);
            }

            public string GetComponentName()
            {
                return TypeDescriptor.GetComponentName(this, true);
            }

            public object GetPropertyOwner(PropertyDescriptor pd)
            {
                return this;
            }

            public AttributeCollection GetAttributes()
            {
                return TypeDescriptor.GetAttributes(this, true);
            }

            public PropertyDescriptorCollection GetProperties(Attribute[] attributes)
            {
                PropertyDescriptorCollection coll = TypeDescriptor.GetProperties(this, attributes, true);

                List<PropertyDescriptor> props = new List<PropertyDescriptor>();

                string engine = Engine.ToLowerInvariant();
                bool engineIsMyIsam = engine == "myisam";

                foreach (PropertyDescriptor pd in coll)
                {
                    if (!pd.IsBrowsable) continue;

                    if (pd.Name == "DataDirectory" || pd.Name == "IndexDirectory")
                    {
                        CustomPropertyDescriptor newPd = new CustomPropertyDescriptor(pd, typeof(TableProperties));
                        newPd.SetReadOnly(!engineIsMyIsam);
                        props.Add(newPd);
                    }
                    else if ((pd.Name == "DelayKeyWrite" || pd.Name == "CheckSum" || pd.Name == "PackKeys") &&
                            !engineIsMyIsam)
                    {
                    }
                    else if (pd.Name == "InsertMethod" && engine != "mrg_myisam") { }
                    else
                        props.Add(pd);
                }
                return new PropertyDescriptorCollection(props.ToArray());
            }

            PropertyDescriptorCollection System.ComponentModel.ICustomTypeDescriptor.GetProperties()
            {
                return TypeDescriptor.GetProperties(this, true);
            }

            public object GetEditor(Type editorBaseType)
            {
                return TypeDescriptor.GetEditor(this, editorBaseType, true);
            }

            public PropertyDescriptor GetDefaultProperty()
            {
                return TypeDescriptor.GetDefaultProperty(this, true);
            }

            public EventDescriptor GetDefaultEvent()
            {
                return TypeDescriptor.GetDefaultEvent(this, true);
            }

            public string GetClassName()
            {
                return TypeDescriptor.GetClassName(this, true);
            }

            #endregion
        }

        private TableProperties current = new TableProperties();
        private TableProperties old = new TableProperties();
        private TableNode owningNode;

		public Table(TableNode node, DataRow row, DataTable columns)
		{
            owningNode = node;
            IsNew = row == null;

            Columns = new TablePartCollection<Column>();
            Indexes = new TablePartCollection<Index>();
            ForeignKeys = new TablePartCollection<ForeignKey>();

            // set some defaults that may be overridden with actual table data
            current.Engine = node.DefaultStorageEngine;
            current.PackKeys = PackKeysMethod.Default;
            current.Schema = node.Database;

            if (row != null)
                current = new TableProperties(row);

            if (columns != null)
              ParseColumns(columns);
            if (!IsNew)
            {
//                LoadIndexes();
  //              LoadForeignKeys();
            }
            old = current;
        }

        internal bool IsNew { get; private set; }
        public TablePartCollection<Column> Columns { get; private set; }
        public TablePartCollection<Index> Indexes { get; private set; }
        public TablePartCollection<ForeignKey> ForeignKeys { get; private set; }

        #region Mirrored Properties

        public string Name
        {
            get { return current.Name; }
            set { current.Name = value; }
        }

        public string CharacterSet
        {
            get { return current.CharacterSet; }
        }

        public string Engine
        {
            get { return current.Engine; }
        }

        #endregion

        public TableProperties Properties { get { return current; } }

        internal TableNode OwningNode
        {
            get { return owningNode; }
        }

        internal bool SupportsFK
        {
            get
            {
                string engine = current.Engine.ToLowerInvariant();
                return engine == "innodb" || engine == "falcon";
            }
        }

        #region ShouldSerializeMethods

        bool ShouldSerializeName() { return false; }
        bool ShouldSerializeSchema() { return false; }
        bool ShouldSerializeComment() { return false; }
        bool ShouldSerializeCharacterSet() { return false; }
        bool ShouldSerializeCollation() { return false; }
        bool ShouldSerializeAutoInc() { return false; }
        bool ShouldSerializeEngine() { return false; }
        bool ShouldSerializeDataDirectory() { return false; }
        bool ShouldSerializeIndexDirectory() { return false; }
        bool ShouldSerializeRowFormat() { return false; }
        bool ShouldSerializeCheckSum() { return false; }
        bool ShouldSerializeAvgRowLength() { return false; }
        bool ShouldSerializeMinRows() { return false; }
        bool ShouldSerializeMaxRows() { return false; }
        bool ShouldSerializePackKeys() { return false; }
        bool ShouldSerializeInsertMethod() { return false; }

        #endregion

        public bool HasChanges()
        {
            return false;
        }

        public void DeleteKey(string keyName)
        {
            for (int i = Indexes.Count - 1; i >= 0; i--)
            {
                if ((keyName != null && Indexes[i].Name == keyName) ||
                    (keyName == null && Indexes[i].IsPrimary))
                {
                    Indexes.Delete(i);
                    break;
                }
            }
        }

        public Index CreateIndexWithUniqueName(bool primary)
        {
            Index newIndex = new Index(this);
            newIndex.IsPrimary = primary;
            string baseName = String.Format("{0}_{1}", primary ? "PK" : "IX", Name);
            string name = baseName;
            int uniqueIndex = 0;
            while (KeyExists(name))
                name = String.Format("{0}_{1}", baseName, ++uniqueIndex);
            newIndex.Name = name;
            return newIndex;
        }

        public List<string> GetColumnNames()
        {
            string sql = @"SELECT column_name FROM INFORMATION_SCHEMA.COLUMNS WHERE 
                TABLE_SCHEMA='{0}' AND TABLE_NAME='{1}'";
            DataTable dt = owningNode.GetDataTable(String.Format(sql, owningNode.Database, Name));
            List<string> cols =new List<string>();
            foreach (DataRow row in dt.Rows)
                cols.Add(row[0].ToString());
            return cols;
        }

        public string GetSql()
        {
            StringBuilder sql = new StringBuilder();
/*            if (IsNew)
                sql.AppendFormat("CREATE TABLE `{0}` (", Name);
            else
                sql.AppendFormat("ALTER TABLE `{0}` ", OldTable.Name);

            string[] parts = new string[3];
            parts[0] = Columns.GetSql(IsNew);
            parts[1] = Indexes.GetSql(IsNew);
            parts[2] = ForeignKeys.GetSql(IsNew);
            string delimiter = "";
            foreach (string s in parts)
            {
                if (!String.IsNullOrEmpty(s))
                {
                    sql.AppendFormat("{0}{1}", delimiter, s);
                    delimiter = ", ";
                }
            }
            if (IsNew)
                sql.Append(")");
            sql.Append(GetTableOptionSql(true));*/
            return sql.ToString();
        }

/*        public override bool Equals(object obj)
        {
            if (!(obj is Table)) return false;
            Table t = obj as Table;

            if (!current.Equals(old)) return false;
            // first compare our top level properties
            if (!ObjectComparer.AreEqual(this, t))
                return false;
//            if (Columns.HasChanges()) return false;
  //          if (Indexes.HasChanges()) return false;
    //        if (ForeignKeys.HasChanges()) return false;
            return true;
        }
        */
        #region Private methods

        private bool KeyExists(string keyName)
        {
            foreach (Index i in Indexes)
                if (String.Compare(i.Name, keyName, true) == 0) return true;
            return false;
        }

        private void ParseColumns(DataTable columnData)
        {
            foreach (DataRow row in columnData.Rows)
            {
                Column c = new Column(row);
                c.OwningTable = this;
                Columns.Add(c);
            }
        }

/*        private void LoadIndexes()
        {
            string[] restrictions = new string[4] { null, owningNode.Database, Name, null };
            DataTable dt = owningNode.GetSchema("Indexes", restrictions);
            foreach (DataRow row in dt.Rows)
            {
                Index i = new Index(this, row);
                Indexes.Add(i);
            }
        }

        private void LoadForeignKeys()
        {
            string[] restrictions = new string[4] { null, owningNode.Database, Name, null };
            DataTable dt = owningNode.GetSchema("Foreign Keys", restrictions);
            foreach (DataRow row in dt.Rows)
            {
                ForeignKey key = new ForeignKey(this, row);
                ForeignKeys.Add(key);
            }
        }*/

        private string GetTableOptionSql(bool newTable)
        {
            List<string> options = new List<string>();
            StringBuilder sql = new StringBuilder(" ");

            if (!newTable)
            {
                if (current.Name != old.Name)
                    options.Add(String.Format("RENAME TO `{0}` ", current.Name));
            }
            if (current.AutoInc != old.AutoInc)
                options.Add(String.Format("AUTO_INCREMENT={0}", current.AutoInc));
            if (current.AvgRowLength != old.AvgRowLength)
                options.Add(String.Format("AVG_ROW_LENGTH={0}", current.AvgRowLength));
            if (CharacterSet != old.CharacterSet)
                options.Add(String.IsNullOrEmpty(CharacterSet) ? "DEFAULT CHARACTER SET" :
                    String.Format("CHARACTER SET='{0}'", CharacterSet));
            if (current.Collation != old.Collation)
                options.Add(String.IsNullOrEmpty(current.Collation) ? "DEFAULT COLLATE" :
                    String.Format("COLLATE='{0}'", current.Collation));
            if (current.CheckSum != old.CheckSum)
                options.Add(String.Format("CHECKSUM={0}", current.CheckSum ? 1 : 0));
            if (current.Comment != old.Comment)
                options.Add(String.Format("COMMENT='{0}'", current.Comment));
            if (Engine != old.Engine)
                options.Add(String.Format("ENGINE={0}", Engine));
            if (current.InsertMethod != old.InsertMethod)
                options.Add(String.Format("INSERT_METHOD={0}", current.InsertMethod.ToString()));
            if (current.MaxRows != old.MaxRows)
                options.Add(String.Format("MAX_ROWS={0}", current.MaxRows));
            if (current.MinRows != old.MinRows)
                options.Add(String.Format("MIN_ROWS={0}", current.MinRows));
            if (current.PackKeys != old.PackKeys)
                options.Add(String.Format("PACK_KEYS={0}", current.PackKeys.ToString()));
            if (current.RowFormat != old.RowFormat)
                options.Add(String.Format("ROW_FORMAT={0}", current.RowFormat.ToString()));
            if (current.DataDirectory != old.DataDirectory)
                options.Add(String.Format("DATA DIRECTORY='{0}' ", current.DataDirectory));
            if (current.IndexDirectory != old.IndexDirectory)
                options.Add(String.Format("INDEX DIRECTORY='{0}' ", current.IndexDirectory));

            string delimiter = "";
            foreach (string option in options)
            {
                sql.AppendFormat("{0}{1}", delimiter, option);
                delimiter = ",\r\n";
            }
            return sql.ToString();
        }

        #endregion
    }
}
